"""empty message

Revision ID: 9bbfdcf8bab0
Revises: fa1760d35710
Create Date: 2020-07-19 01:04:23.442598

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '9bbfdcf8bab0'
down_revision = 'fa1760d35710'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('experiment_runs', sa.Column('random_seed', sa.Integer(), nullable=True), schema='triage_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('experiment_runs', 'random_seed', schema='triage_metadata')
    # ### end Alembic commands ###
