"""empty message

Revision ID: 4ae804cc0977
Revises: 9bbfdcf8bab0
Create Date: 2020-07-19 01:35:54.419099

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '4ae804cc0977'
down_revision = '9bbfdcf8bab0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('models', sa.Column('built_in_experiment_run', sa.Integer(), nullable=True), schema='triage_metadata')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('models', 'built_in_experiment_run', schema='triage_metadata')
 # ### end Alembic commands ###
